# = ChangeLogReader
# ChangeLogReader.pm の Ruby 版
#
require 'date'

class Date
  def ym
    "#{self.year}-#{self.month}"
  end
  def md
    "#{self.month}-#{self.day}"
  end
end

class ChangeLogReader

  REG_DATE = /^(\d{4}-\d{2}-\d{2})/

  def initialize(ignore_private = nil, stop_date = nil)
    @ignore_private = ignore_private
    @entries = []
    @data = ""   # original data
    @category = Hash.new
    @stat_month = Hash.new(0)
    @stat_same_date = {} # {md} = [year, year....]
    @month_list = {} # @month_list{ym} = [ ym が示す月の entry]
    @stop_date = stop_date
  end
  
  def debug_print()
    
  end

  # ChangeLog の読み込み
  def read(file)
    @data += open(file).read
  end

  attr_reader :entries, :stat_month, :stat_same_date, :category,
  :month_list, :stop_date, :stat_category, :inside_ref_from, :ignore_private

  def parse
    split_into_entry

    @entries.each do |entry|
      if @month_list[entry.date.ym].nil?
        @month_list[entry.date.ym] = [entry]
      else
        @month_list[entry.date.ym].push(entry)
      end
    end

    update_stat
  end

  private

  def split_into_entry
    entry_lines = []

    @data.each do |line|
      line.chomp!
      if REG_DATE =~ line
        unless entry_lines.empty?
          @entries.push(Entry.new(entry_lines, @ignore_private)) 
        end
        entry_lines = []
      elsif /^\t?__DATA__.*$/ =~ line
        break
      end
      entry_lines.push(line)
    end
    unless entry_lines.empty?
      @entries.push(Entry.new(entry_lines, @ignore_private)) 
    end
  end

  # 統計情報の更新
  def update_stat
    @inside_ref_from = {}
    @entries.each do |entry|
      @stat_month[entry.date.ym] += entry.items.size
      entry.items.each do |item|
        a = item.content.scan(/\[(\d{4}-\d{2}-\d{2}(?:-\d+)?)\]/)
        if a
          a.flatten.each do |i|
            @inside_ref_from[i] = "[#{item.date_with_id}]"
          end
        end

        if @inside_ref_from[item.date_with_id]
          item.inside_ref_from = @inside_ref_from[item.date_with_id]
        end

        md = "#{item.date.md}"
        if @stat_same_date[md].nil?
          @stat_same_date[md] = [item.date.year]
        else
          @stat_same_date[md].push(item.date.year)
        end
        check_category(item)
      end
    end
  end

  # 各 category の item 数を更新
  def check_category(item)
    return if item.category.nil?

    item.category.each do |cat|
      if @category[cat].nil?
        @category[cat] = [item]
      else
        @category[cat].push(item)
      end
    end
    @stat_category = @category.map{|x| x.size}
  end

  # entry を扱う
  class Entry
    REG_ENTRY_HEADER_DATE = /^\d{4}-\d{2}-\d{2}/
    REG_ENTRY_HEADER = /^((\d{4})-(\d{2})-(\d{2}))(?:\s*\(.{3}\))?\s*(.+)$/

    def initialize(lines, ignore_private = nil)
      @ignore_private = ignore_private
      @entry = lines # original data
      @date = nil    # Date object
#      @item_id = ""     # 200511041
      @message_top = nil
      @message_bottom = nil
      @items = []
      @items_without_private_item = []
      @header = ""

      parse
    end

    attr_reader :date, :items, :items_without_private_item,
    :header

    private

    def parse
      @header = @entry.shift

      return unless REG_ENTRY_HEADER_DATE =~ @header
      date, year, month, day, user = @header.scan(REG_ENTRY_HEADER)[0]
      @date = Date.parse(date)
      @user = user.gsub(/</, "&lt;").gsub(/>/, "^gt;")

      split_into_item

      if @items.empty?
        return nil
      else
        return self
      end
    end
    
    # item に分割
    def split_into_item
      puts "entering split_into_item" if $DEBUG
      item_lines = []
      items = []
      
      @entry.each do |line|
        if /^( {8}| {0,7}\t|)\* / =~ line
          item_lines.pop if /^\s*$/ =~ item_lines[-1]
          items.push(item_lines.map{|x| x.sub(/^\t/, "")}) unless
            (item_lines.empty? or /^\s*$/ =~ item_lines[0] or item_lines.nil?)
          item_lines = []
        end
        item_lines.push(line)
      end
      
      item_lines.pop if /^\s*$/ =~ item_lines[-1]
      items.push(item_lines.map{|x| x.sub(/^\t/, "")}) unless
        (item_lines.empty? or /^\s*$/ =~ item_lines[0] or item_lines.nil?)

      # reverse する必要ある？
      items = items.reverse.each_with_index do |item, idx|
        if item.nil?
          exit
        else
          item = Item.new(item, @date, @user, idx)
          store_item(item) unless @ignore_private and item.private?
        end
      end
    end

    # item の種類に応じて store
    def store_item(item)
      if item.message_top?
        @message_top = item
      elsif item.message_bottom?
        @message_bottom = item
      end

      # ここで private をわける必要は？
      @items.push(item)
      @items_without_private_item.push(item) unless item.private?
    end

    class Item
      def initialize(item, date, user, id)
        @data = item
        @date = date
        @user = user
        @id = (id + 1).to_s # item ID : Y in XXXX-XX-XX-Y
        @date_with_id = @date.to_s + "-" + @id
        @category = nil
        @header = ""
        @content = ""
        @private = false
        @message_top = false
        @message_bottom = false
#        @inside_link = []
        @inside_ref_from = []

        parse
      end
      
      def private?
        @private
      end
      
      def message_top?
        @message_top
      end

      def message_bottom?
        @message_bottom
      end

      attr_reader :category, :user, :date, :id, :date_with_id
      attr_accessor :content, :header, :inside_ref_from

      private

      def parse
#        item_header = @data.shift.sub(/:(\s.*?)$/, ":")
        ih = @data.shift
        if /\*.+:\s*$/ =~ ih
          item_header, rest = ih.scan(/\s*(.+):(\s*)$/)[0]
        else
          item_header = ih
        end

        item_header.strip! if /\S/ =~ item_header

        if rest
          rest = rest.sub(/^ +/, "")
        else
          rest = ""
        end

        parse_item_header(item_header)
        parse_item_content(rest)
        @header = item_header
      end

      # Processing item header
      # private, pragma, category の check
      def parse_item_header(ih)
        ih.sub!(/^\* /, "")

        case ih
        when /^\s*p:/
          @private = true
        when /^message-top:/
          @message_top = true
        when /^message-bottom:/
          @message_bottom = true
        end

        # Triming trailing spaces and ":"
        ih.gsub!(/(:|\s*)$/, "")

        if ih.match(/\s*\[(.+)\]$/) # category
          @category = $1.split(/\s*\]\s*\[\s*/)
        end
      end

      # item header の行で
      #  * hoge: foo bar
      # のように本文が始まっている場合の処理
      #
      # 最終行の "\n" が必要
      def parse_item_content(rest)
        if rest.nil? or rest.empty?
          @content = @data.map{|x| x.gsub(/\r/, "")}.join("\n") + "\n"
        else
          @content = ([rest] + @data).map{|x| x.gsub(/\r/, "")}.join("\n") + "\n"
        end
      end
    end
  end
end
